import sys
from pathlib import Path
from typing import Literal, Optional
import json

from datargs import argsclass, parse

from .data import QDXV1QCInput, run_convert
from .api import PrepTypes, QDXProvider


@argsclass(description="QDX CLI")
class QDXArgs:
    url: str
    access_token: str
    # input: str = arg(positional=True, help="QDXV1 Input json")
    post_quantum_energy: Optional[bool]
    get_quantum_energy: Optional[str]
    delete_proc: Optional[str]
    tag_proc: Optional[str]
    untag_proc: Optional[str]
    pdb_to_complex: Optional[Path]
    fragment_complex: Optional[int]
    tags: Optional[list[str]]
    convert: Optional[Path]
    prepare_protein: Optional[PrepTypes]
    prepare_ligand: Optional[PrepTypes]
    combine_complexes: Optional[Path]
    poll: Optional[bool]
    direction: Optional[Literal["qdxv12exess", "exess2qdxv1", "qdxcomplex2qdxv1", "qdxcomplex2exess"]]


def main():
    args = parse(QDXArgs)

    provider = QDXProvider(args.url, args.access_token)

    if args.delete_proc:
        print(provider.delete_proc(args.delete_proc))
    elif args.tag_proc:
        print(provider.tag_proc(args.tag_proc, args.tags or []))
    elif args.untag_proc:
        for tag in args.tags or []:
            print(provider.untag_proc(args.untag_proc, tag))
    elif args.post_quantum_energy:
        print(
            provider.start_quantum_energy_calculation(
                QDXV1QCInput.from_json(sys.stdin.read()), tags=args.tags or []
            )
        )
    elif args.prepare_protein:
        id = provider.start_protein_prep(
            json.load(sys.stdin), prep_for=args.prepare_protein, tags=args.tags or []
        )
        print(json.dumps(provider.poll_proc(id).state["term"]) if args.poll else id)
    elif args.prepare_ligand:
        id = provider.start_ligand_prep(sys.stdin.read(), prep_for=args.prepare_ligand, tags=args.tags or [])
        print(json.dumps(provider.poll_proc(id).state["term"]) if args.poll else id)
    elif args.get_quantum_energy:
        print(provider.get_proc(args.get_quantum_energy).to_json())
    elif args.pdb_to_complex:
        print(json.dumps(provider.pdb_to_complex(args.pdb_to_complex)))
    elif args.combine_complexes:
        print(
            json.dumps(
                provider.combine_complexes(
                    json.loads(args.combine_complexes.read_text()), json.load(sys.stdin)
                )
            )
        )
    elif args.fragment_complex:
        print(json.dumps(provider.fragment_complex(json.load(sys.stdin), args.fragment_complex)))
    elif args.convert and args.direction:
        with open(args.convert) as f:
            print(run_convert(f.read(), args.direction).to_json(indent=2))


if __name__ == "__main__":
    main()



"""
Utilities for creating QC JSON files (https://molssi-qc-schema.readthedocs.io/en/latest/index.html)
"""
from dataclasses import dataclass, field
import json
from pathlib import Path
from typing import Any, Literal, Optional

import dataclasses_json
from tap import Tap

@dataclass
class Vec3f(dataclasses_json.DataClassJsonMixin):
    """Vector for 3d coordinates."""

    x: float
    y: float
    z: float


@dataclass
class BindingSite(dataclasses_json.DataClassJsonMixin):
    """Details of a receptor binding site."""

    center: Vec3f
    size: Optional[Vec3f]


@dataclass
class FragAtom(dataclasses_json.DataClassJsonMixin):
    """Atom in a Fragment."""

    index: int  # internal index used for ordering
    element: str
    coords: list[float]
    neighbors: list[str]  # list of indexes for determining broken bonds
    residue: int


@dataclass
class Fragment(dataclasses_json.DataClassJsonMixin):
    """Represents sub-sections of a contigous molecular chain."""

    atoms: list[FragAtom]
    residues: list[tuple[int, Any]]

    first_atom_coords: str
    final_atom_coords: str

    charge: float


@dataclass
class Section(dataclasses_json.DataClassJsonMixin):
    """Represents different complete contiguous molecular chains - ie an unbroken protein chain."""

    atoms: list[FragAtom]
    residues: list[tuple[int, Any]]


QCMethods = Literal["RHF", "RIMP2"]

QCBasisSet = Literal[
    "STO-2G",
    "STO-3G",
    "STO-4G",
    "STO-5G",
    "STO-6G",
    "3-21G",
    "4-31G",
    "5-21G",
    "6-21G",
    "6-31G",
    "6-31G*",
    "6-31G**",
    "6-31+G",
    "6-31+G*",
    "6-31+G**",
    "6-31++G",
    "6-31++G*",
    "6-31++G**",
    "6-31G(2df,p)",
    "6-31G(3df,3pd)",
    "6-31G(d,p)",
    "6-311G",
    "6-311G*",
    "6-311G**",
    "6-311+G",
    "6-311+G*",
    "6-311+G**",
    "6-311++G",
    "6-311++G*",
    "6-311++G**",
    "6-311G(d,p)",
    "6-311G(2df,2pd)",
    "6-311+G(2d,p)",
    "6-311++G(2d,2p)",
    "6-311++G(3df,3pd)",
    "PCSeg-0",
    "PCSeg-1",
    "PCSeg-2",
    "cc-pVDZ-RIFIT",
    "aug-cc-pVDZ-RIFIT",
    "cc-pVTZ-RIFIT",
    "aug-cc-pVTZ-RIFIT",
    "cc-pVDZ",
    "aug-cc-pVDZ",
    "cc-pVTZ",
    "aug-cc-pVTZ",
]


class DataClassJsonMixin(dataclasses_json.DataClassJsonMixin):
    """Override dataclass mixin so that we don't have `"property": null,`s in our output"""

    dataclass_json_config = dataclasses_json.config(  # type: ignore
        undefined=dataclasses_json.Undefined.EXCLUDE,
        exclude=lambda f: f is None,  # type: ignore
    )["dataclasses_json"]


@dataclass
class QCJSONFragments(DataClassJsonMixin):
    """ "fragments" in qc json"""

    fragid: list[int] = field(default_factory=list)
    nfrag: int = 0
    broken_bonds: list[int] = field(default_factory=list)
    fragment_charges: list[int] = field(default_factory=list)
    frag_size: list[int] = field(default_factory=list)


@dataclass
class QCMol(DataClassJsonMixin):
    """ "molecule" in qc json"""

    geometry: list[float] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)
    fragments: QCJSONFragments = QCJSONFragments()


@dataclass
class QCModel(DataClassJsonMixin):
    """ "model" in qc json"""

    method: QCMethods = "RHF"
    fragmentation: bool = False
    basis: QCBasisSet = "6-31G*"
    aux_basis: str = "cc-pVDZ-RIFIT"


@dataclass
class QCSCF(DataClassJsonMixin):
    """ "keywords.scf" in qc json"""

    niter: int = 40
    ndiis: int = 12
    dele: float = 1e-5
    rmsd: float = 1e-6
    dynamic_threshold: int = 10
    debug: bool | None = None
    convergence_metric: str = "diis"


@dataclass
class QCFrag(DataClassJsonMixin):
    """ "keywords.frag" in qc json"""

    reference_monomer: Optional[int] = None
    method: str = "MBE"
    level: int = 2
    ngpus_per_group: int = 4
    dimer_cutoff: int = 40
    dimer_mp2_cutoff: int = 40
    trimer_mp2_cutoff: int = 20
    lattice_energy_calc: bool = False
    trimer_cutoff: int = 30


@dataclass
class QCRIMP2(DataClassJsonMixin):
    """ "keywords.rimp2" in qc json"""

    box_dim: int = 15


@dataclass
class QCKeywords(DataClassJsonMixin):
    """ "keywords" in qc json"""

    scf: QCSCF = QCSCF()
    frag: QCFrag = QCFrag()
    rimp2: QCRIMP2 = QCRIMP2()


@dataclass
class QCInput(DataClassJsonMixin):
    """Final qc json input file"""

    molecule: QCMol = QCMol()
    model: QCModel = QCModel()
    keywords: QCKeywords = QCKeywords()
    driver: str = "energy"


@dataclass
class QDXV1QCMol(DataClassJsonMixin):
    """ "molecule" in qc json"""

    geometry: list[float] = field(default_factory=list)
    symbols: list[str]= field(default_factory=list)
    fragments: Optional[list[list[int]]] = field(default_factory=list)
    connectivity: Optional[list[tuple[int, int, int]]] = field(default_factory=list)
    fragment_charges: Optional[list[int]] = field(default_factory=list)


@dataclass
class QDXV1QCInput(DataClassJsonMixin):
    """Final qc json input file"""

    molecule: QDXV1QCMol = QDXV1QCMol()
    model: QCModel = QCModel()
    keywords: QCKeywords = QCKeywords()
    driver: str = "energy"


@dataclass
class QCResults(DataClassJsonMixin):
    """Energy results parsed from an EXESS run"""

    calculated_fragments: int
    date: str
    final_hf_energy: float
    final_mp_energy: float
    nmer_distances: dict[str, float]
    nmer_hf_energies: dict[str, float]
    nmer_mp2_ss_energies: dict[str, float]
    nmer_mp2_os_energies: dict[str, float]
    input_file: str
    log_name: str
    num_atoms: int
    un_converged_fragments: int
    run_time: float
    xyzs: list[str]


@dataclass
class QDXV1Energy(DataClassJsonMixin):
    energy_type: Literal["lattice", "SPE"] | None  # Whether the calculation is lattice against a ref. monomer
    hf_total: float  # Final single-point energy (HF)
    mp_ss_total: float  # Final single-point energy correction (MP2 same-spin)
    mp_os_total: float  # Final single-point energy correction (MP2 opposite-spin)

    hf: list[float]  # HF energy per fragment group
    mp_ss: list[float]  # MP2 same-spin corrections per fragment group
    mp_os: list[float]  # MP2 opposite-spin corrections per fragment group
    n_mers: list[list[int]]  # Fragment groups
    n_mer_distances: list[float] | None  # Distances between fragments


TupleAtom = tuple[str, list[float]]


def qc_get_fragment_atoms(m: QCMol) -> list[list[TupleAtom]]:
    frags: list[list[TupleAtom]] = []
    for aidx, f in enumerate(m.fragments.fragid):
        while len(frags) < f:
            frags.append([])
        frags[f - 1].append((m.symbols[aidx], m.geometry[aidx * 3 : aidx * 3 + 3]))
    return frags


def exess_qc_mol_to_qdx_v1_mol(mol: QCMol) -> QDXV1QCMol:
    """Turns a Exess/Hermes molecule input object into a compliant qc input object"""
    frags: list[list[int]] = []
    for aidx, f in enumerate(mol.fragments.fragid):
        while len(frags) < f:
            frags.append([])
        frags[f - 1].append(aidx)
    bpairs = zip(*[iter(mol.fragments.broken_bonds)] * 2, strict=True)
    connectivity: list[tuple[int, int, int]] = [(x - 1, y - 1, 1) for x, y in bpairs]  # type: ignore
    return QDXV1QCMol(
        mol.geometry,
        mol.symbols,
        frags,
        connectivity,
        mol.fragments.fragment_charges,
    )


def qdx_v1_qc_mol_to_exess_mol(mol: QDXV1QCMol) -> QCMol:
    """Turns a compliant molecule input object into a Exess/Hermes qc input object.

    NOTE: connectivity is taken as broken bonds only!
    """
    fragids = list(range(len(mol.symbols)))
    if mol.fragments:
        for i, frag in enumerate(mol.fragments):
            for j in frag:
                fragids[j] = i + 1
                # fragids.append(i + 1)

    broken_bonds = []

    if mol.connectivity:
        for a1, a2, _ in mol.connectivity:
            broken_bonds.extend([a1 + 1, a2 + 1])

    return QCMol(
        mol.geometry,
        mol.symbols,
        QCJSONFragments(
            fragids,
            len(mol.fragments or []),
            broken_bonds,
            mol.fragment_charges or [],
            [len(x) for x in mol.fragments or []],
        ),
    )


def qcresults_to_qdx_v1_energy(results: QCResults):
    n_mers = []
    hf = []
    mp_os = []
    mp_ss = []
    nmer_distances = []
    energy_type = "lattice"
    reference = None

    mp_os_total = 0
    mp_ss_total = 0

    for (k, v) in results.nmer_hf_energies.items():
        mers = list(map(int, k.split(":")))
        if len(mers) > 1:
            if reference:
                if reference != mers[0]:
                    energy_type = "SPE"
            else:
                reference = mers[0]

        n_mers.append(mers)
        hf.append(v)
        mp_os.append(results.nmer_mp2_os_energies.get(k))
        mp_ss.append(results.nmer_mp2_ss_energies.get(k))
        nmer_distances.append(results.nmer_distances.get(k))
        if len(mers) > 1:
            mp_os_total += results.nmer_mp2_os_energies.get(k) or 0
            mp_ss_total += results.nmer_mp2_ss_energies.get(k) or 0

    if energy_type != "lattice":
        mp_os_total = sum(results.nmer_mp2_os_energies.values())
        mp_ss_total = sum(results.nmer_mp2_ss_energies.values())

    return QDXV1Energy(
        energy_type=energy_type,
        hf_total=results.final_hf_energy,
        mp_os_total=mp_os_total,
        mp_ss_total=mp_ss_total,
        hf=hf,
        mp_os=mp_os,
        mp_ss=mp_ss,
        n_mers=n_mers,
        n_mer_distances=nmer_distances,
    )


class QCJsonArgs(Tap):
    input: Path
    direction: Literal["qdxv12exess", "exess2qdxv1", "qdxcomplex2qdxv1", "qdxcomplex2exess"]


def run_convert(data: str, direction: Literal["qdxv12exess", "exess2qdxv1", "qdxcomplex2qdxv1", "qdxcomplex2exess"]):
    """Convert between multiple exess/hermes formats.
    """
    match direction:
        case "qdxcomplex2qdxv1":
            input_json = json.loads(data)
            input = QDXV1QCMol.from_dict(input_json["topology"])
            return QDXV1QCInput(input, QCModel(), QCKeywords(QCSCF(), QCFrag(), QCRIMP2()))
        case "qdxcomplex2exess":
            input_json = json.loads(data)
            input = QDXV1QCMol.from_dict(input_json["topology"])
            n = qdx_v1_qc_mol_to_exess_mol(input)
            return QCInput(n, QCModel(), QCKeywords(QCSCF(), QCFrag(), QCRIMP2()))
        case "exess2qdxv1":
            input = QCInput.from_json(data)
            n = exess_qc_mol_to_qdx_v1_mol(input.molecule)
            return QDXV1QCInput(n, input.model, input.keywords, input.driver)
        case "qdxv12exess":
            input = QDXV1QCInput.from_json(data)
            n = qdx_v1_qc_mol_to_exess_mol(input.molecule)
            return QCInput(n, input.model, input.keywords, input.driver)


def main():
    """Entrypoint to qdx format conversion helpers
    Parse arguments and write output to files
    """
    args = QCJsonArgs().parse_args()
    with open(args.input) as f:
        print(run_convert(f.read(), args.direction).to_json(indent=2))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qdx']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.5.7,<0.6.0',
 'datargs>=0.11.0,<0.12.0',
 'gql[requests]>=3.4.0,<4.0.0',
 'typed-argument-parser>=1.7.2,<2.0.0']

setup_kwargs = {
    'name': 'qdx',
    'version': '0.2.0',
    'description': '',
    'long_description': '# QDX-py: Python SDK for the QDX API\n\nThis package exposes a simple provider and CLI for the different tools exposed by the QDX GraphQL API.\n\n## Usage\n\n### As a library\n\n``` python\nfrom qdx.api import QDXProvider\n\nURL = "url to the qdx api"\nTOKEN = "your qdx access token"\n\nprovider = QDXProvider(URL, TOKEN)\n\ninput = # Some QDXV1QCInput\n\ntask_id = provider.start_quantum_energy_calculation(input) # will return a TaskId - reference to the job\n\ntask = provider.get_quantum_energy_calculation(task_id) # will return a task, with its status, progress, and result if completed\n\n# NOTE: tasks may take a while to run, so you will have to poll the task until it\'s done\n```\n\n\n### As a CLI\n\n``` sh\n# All cli calls have these standard arguments, referred to as … in future examples\nqdx --url QDX_API_URL --access-token QDX_ACCESS_TOKEN\n\n# Post a hermes job, returning a task id\n… --post-quantum-energy < ./path_to_qdxv1_input.json\n\n# Retrieve the hermes job, or its progress\n… --get-proc TASK_ID\n\n## Other functions\n# Return a qdx complex json object and save it as complex.json\n… --pdb-to-complex PATH_TO_PDB_FILE > complex.json\n\n# Prepare a protein for quauntum energy calculation\n… --prepare-protein simulation --poll < ./complex.json > prepped_protein_complex.json\n\n# Fragment a qdx complex json object\n… --fragment-complex [MIN_STEPS_ALONG_PROTEIN_BACKBONE_BEFORE_CUTTING_AT_C-C_BOND] < prepped_protein_complex.json > fragmented_protein_complex.json\n\n# Prepare a ligand for quauntum energy calculation\n… --prepare-ligand simulation --poll < ./path_to_ligand.sdf > prepped_ligand_complex.json\n\n# Combine protein and ligand complexes for simulation\n\n… --combine-complexes ./prepped_protein_complex.json < ./prepped_ligand_complex.sdf > protein_ligand_complex.json\n\n# Convert a qdx complex into a qdx input file\n… --convert ./protein_ligand_complex.json --direction qdxcomplex2qdxv1 > qdx_input.json\n\n# Convert a qdx complex into a exess input file\n… --convert ./protein_ligand_complex.json --direction qdxcomplex2exess > exess_input.json\n\n# Convert a qdx input file into an exess input file\n… --convert qdx_input.json --direction qdxv12exess > exess_input.json\n```\n\n',
    'author': 'Ryan Swart',
    'author_email': 'ryan@talosystems.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, Literal, Optional, TypeVar

import dataclasses_json
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from .data import BindingSite, QDXV1Energy, QDXV1QCInput, QDXV1QCMol

ProcId = str


quantum_energy_mutation = gql(
    """
    mutation calculate_quantum_energy($input: CalculateQuantumEnergyArgs, $tags: [String]) {
        calculateQuantumEnergy(input: $input, tags: $tags)
    }
    """
)

delete_proc_mutation = gql(
    """
    mutation delete_proc($procId: ProcId) {
        deleteProc(procId: $procId)
    }
    """
)

tag_proc_mutation = gql(
    """
    mutation tag_proc($procId: ProcId, $tags: [String]) {
        tagProc(procId: $procId, tags: $tags)
    }
    """
)

untag_proc_mutation = gql(
    """
    mutation untag_proc($procId: ProcId, $tag: String) {
        untagProc(procId: $procId, tag: $tag)
    }
    """
)

pdb_to_complex_query = gql(
    """
    query pdb_to_complex($pdb: String!, $withConnectivity: Boolean!) {
        pdbToComplex(pdb: $pdb, withConnectivity: $withConnectivity) {
            topology {
                symbols
                geometry
                connectivity
                fragments
                fragment_charges
                atom_charges
            }
            amino_acids
            amino_acid_atoms
            atoms
            atom_alts
            atom_alt_charges
            residues
            residue_atoms
            subunits
        }
    }
    """
)

combine_complexes_query = gql(
    """
    query combine_complexes($complex: ComplexInput!, $complex_b: ComplexInput!) {
        combineComplexes(complex: $complex, complexB: $complex_b) {
            topology {
                symbols
                geometry
                connectivity
                fragments
                fragment_charges
                atom_charges
            }
            amino_acids
            amino_acid_atoms
            atoms
            atom_alts
            atom_alt_charges
            residues
            residue_atoms
            subunits
        }
    }
    """
)


fragment_complex_query = gql(
    """
    query fragment($complex: ComplexInput!, $backboneSteps: Int!, $skipCharges: Boolean!) {
        fragmentComplex(complex: $complex, backboneSteps: $backboneSteps, skipExplicitCharges: $skipCharges) {
            topology {
                symbols
                geometry
                connectivity
                fragments
                fragment_charges
                atom_charges
            }
            amino_acids
            amino_acid_atoms
            atoms
            atom_alts
            atom_alt_charges
            residues
            residue_atoms
            subunits
        }
    }
    """
)

complex_to_xyz_query = gql(
    """
    query complex_xyz($complex: ComplexInput!) {
        complexToXyz(complex: $complex)
    }
    """
)

complex_to_xyz_fragments_query = gql(
    """
    query complex_xyz($complex: ComplexInput!) {
        complexToXyzFragments(complex: $complex)
    }
    """
)

dock_mutation = gql(
    """
    mutation dock($protein: ComplexInput!, $sdf: String!, $binding_site: BindingSiteInput!, $tags: [String]){
        dock(input: { sdf: { string: $sdf }, complex: $protein, binding_site: $binding_site }, tags: $tags)
    }
    """
)

prepare_ligand_mutation = gql(
    """
    mutation prep_lig($sdf: String!, $tags: [String], $prep_for: PrepTypes!){
        prepareLigand(input: { sdf: { string: $sdf }, prep_for: $prep_for }, tags: $tags)
    }
    """
)

prepare_protein_mutation = gql(
    """
    mutation prep_prot($input: ComplexInput!, $tags: [String], $prep_for: PrepTypes!) {
        prepareProtein(input: { complex: $input, prep_for: $prep_for }, tags: $tags)
    }
    """
)


proc_query = gql(
    """query s($input: String) {
  proc(procId: $input) {
    id
    completed_at
    created_at
    deleted_at
    runtime
    queued_at
    scheduled_at
    tags
    data {
      ... on Dock {
        output {
            complex {
                topology {
                    symbols
                    geometry
                    connectivity
                    fragments
                    fragment_charges
                    atom_charges
                }
                amino_acids
                amino_acid_atoms
                atoms
                atom_alts
                atom_alt_charges
                residues
                residue_atoms
                subunits
                }
            score
        }
      }
      ... on PrepareProtein {
        output {
            topology {
                symbols
                geometry
                connectivity
                fragments
                fragment_charges
                atom_charges
            }
            amino_acids
            amino_acid_atoms
            atoms
            atom_alts
            atom_alt_charges
            residues
            residue_atoms
            subunits
        }
      }
      ... on PrepareLigand {
        output {
            topology {
                symbols
                geometry
                connectivity
                fragments
                fragment_charges
                atom_charges
            }
            amino_acids
            amino_acid_atoms
            atoms
            atom_alts
            atom_alt_charges
            residues
            residue_atoms
            subunits
        }
      }
      ... on CalculateQuantumEnergy {
        state {
            current_step
            total_steps
            estimated_completed_at
            stderr_path
            stdout_path
        }
        err {
            msg
        }
        output {
          stderr_path
          stdout_path
          energy {
            energy_type
            n_mers
            n_mer_distances
            hf
            hf_total
            mp_ss
            mp_ss_total
            mp_os
            mp_os_total
          }
        }
      }
    }
  }
}
"""
)


class DataClassJsonMixin(dataclasses_json.DataClassJsonMixin):
    """Override dataclass mixin so that we don't have `"property": null,`s in our output"""

    dataclass_json_config = dataclasses_json.config(  # type: ignore
        undefined=dataclasses_json.Undefined.EXCLUDE,
        exclude=lambda f: f is None,  # type: ignore
    )["dataclasses_json"]


@dataclass
class QDXEnergyResult(DataClassJsonMixin):
    energy: QDXV1Energy
    stderr_path: str
    stdout_path: str


@dataclass
class DDProgress(DataClassJsonMixin):
    """Dataclass representing progress of a workflow step."""

    stderr_path: str
    stdout_path: str
    total_steps: int = 0
    started_at: int = 0
    current_step: int = 0
    estimated_finish: int = 0


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")


class ProcState(Generic[T1, T2, T3], DataClassJsonMixin):
    input: Optional[T1] = None
    state: Optional[T2] = None
    output: Optional[T3] = None


T = TypeVar("T", bound=ProcState)


QDXEnergyState = ProcState[QDXV1QCInput, DDProgress, QDXEnergyResult]

PrepTypes = Literal["docking", "simulation"]


@dataclass
class Proc(Generic[T], DataClassJsonMixin):
    id: str
    completed_at: int | None
    created_at: int
    deleted_at: int | None
    runtime: int
    tags: list[str]
    queued_at: int | None
    scheduled_at: int | None
    data: T
    # error: str | None


class QDXProvider:
    def __init__(self, url: str, access_token: str):
        transport = RequestsHTTPTransport(url=url, headers={"authorization": f"bearer {access_token}"})

        self.client = Client(transport=transport)

    def tag_proc(self, id: ProcId, tags: list[str]):
        response = self.client.execute(tag_proc_mutation, variable_values={"procId": id, "tags": tags})
        taskId = response.get("tagProc")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def untag_proc(self, id: ProcId, tag: str):
        response = self.client.execute(untag_proc_mutation, variable_values={"procId": id, "tag": tag})
        taskId = response.get("untagProc")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def delete_proc(self, id: ProcId):
        response = self.client.execute(delete_proc_mutation, variable_values={"procId": id})

        taskId = response.get("deleteProc")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def start_quantum_energy_calculation(self, input: QDXV1QCInput, tags: list[str]) -> ProcId:
        input.keywords.scf.debug = None
        input_dict = input.to_dict()

        response = self.client.execute(
            quantum_energy_mutation, variable_values={"input": input_dict, "tags": tags}
        )

        taskId = response.get("calculateQuantumEnergy")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def poll_proc(self, id: ProcId, n_retries: int = 10, poll_rate: int = 30) -> Proc:
        """Keep querying proc until it is completed."""
        n_try = 0

        while n_try < n_retries:
            n_try += 1
            response = self.client.execute(proc_query, variable_values={"input": id})
            proc: dict[str, Any] | None = response.get("proc")
            if proc and proc["data"]["output"]:
                deserialized = Proc.from_dict(proc)
                return deserialized
            if proc and proc["data"].get("err"):
                msg = proc["data"]["err"]["msg"]
                raise Exception(f"Proc failed: {msg}")

            time.sleep(poll_rate)


        raise Exception("Proc polling timed out")


    def get_proc(self, id: ProcId) -> Proc:
        response = self.client.execute(proc_query, variable_values={"input": id})
        proc: dict[str, Any] | None = response.get("proc")
        if proc:
            print(proc)
            deserialized = Proc.from_dict(proc)
            return deserialized

        raise Exception("Failed to find task")

    def start_ligand_prep(self, sdf: str, prep_for: PrepTypes, tags: list[str] = []):
        response = self.client.execute(
            prepare_ligand_mutation,
            variable_values={
                "sdf": sdf,
                "tags": tags,
                "prep_for": prep_for.upper()
            },
        )

        taskId = response.get("prepareLigand")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def start_protein_prep(self, complex: Any, prep_for: PrepTypes, tags: list[str] = []):
        response = self.client.execute(
            prepare_protein_mutation,
            variable_values={
                "input": complex,
                "tags": tags,
                "prep_for": prep_for.upper()
            },
        )

        taskId = response.get("prepareProtein")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def start_dock(self, complex: Any, sdf: str, binding_site: BindingSite, tags: list[str] = []):
        response = self.client.execute(
            dock_mutation,
            variable_values={
                "binding_site": binding_site.to_dict(),
                "protein": complex,
                "sdf": sdf,
                "tags": tags,
            },
        )

        taskId = response.get("dock")
        if taskId:
            return ProcId(taskId)
        else:
            raise RuntimeError(response)

    def pdb_to_complex(self, pdb: str | Path, with_connectivity: bool = True) -> Any:
        if isinstance(pdb, str):
            response = self.client.execute(
                pdb_to_complex_query,
                variable_values={"pdb": pdb, "withConnectivity": with_connectivity},
                upload_files=True,
            )
        else:
            with pdb.open(mode="r") as f:
                response = self.client.execute(
                    pdb_to_complex_query,
                    variable_values={"pdb": f.read(), "withConnectivity": with_connectivity},
                )

        complex: dict[str, Any] | None = response.get("pdbToComplex")

        if complex:
            return complex

        raise Exception("Failed convert pdb")

    def fragment_complex(self, complex: Any, backbone_steps: int, skip_explicit_charges: bool = False):
        response = self.client.execute(
            fragment_complex_query,
            variable_values={
                "complex": complex,
                "backboneSteps": backbone_steps,
                "skipCharges": skip_explicit_charges,
            },
        )
        complex = response.get("fragmentComplex")

        if complex:
            return complex

        raise Exception("Failed to fragment complex")

    def combine_complexes(self, complex: Any, complexB: Any):
        response = self.client.execute(
            combine_complexes_query,
            variable_values={
                "complex": complex,
                "complex_b": complexB,
            },
        )
        complex = response.get("combineComplexes")

        if complex:
            return complex

        raise Exception("Failed to fragment complex")
